﻿using Misc;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FlyADBase
{
    /// <summary>
    /// 高级timegrid助手
    /// 使用说明：
    /// XXXPool 只能储存 1min数据
    /// 
    /// 
    /// 获取最后一次运行动作的开始与结束时间
    /// bool ret = mTimeGridAdvHelper.GetLastRunningTime(out DateTime beginTime, out DateTime endTime, out int marker, out DRIVE_MAN_ORDER order);
    /// 
    /// 获取 指定时间范围的 timegrid 数据
    /// var adList = mTimeGridAdvHelper.GetTimeGrid(beginTime, endTime, out DateTime reponse_endTime);
    ///
    /// 获取 指定时间范围的 pos 数据
    /// var posList = mTimeGridAdvHelper.GetPos(reponse_endTime,adList.Count());
    /// 
    /// 获取 指定时间范围的 pos2 数据
    /// var pos2List = mTimeGridAdvHelper.GetPos2(reponse_endTime,adList.Count());
    /// 
    /// 获取 指定时间范围的 输入口 数据
    /// var istatusList = mTimeGridAdvHelper.GetIStatus(reponse_endTime,adList.Count());
    /// 
    /// 转为 grid 数据
    /// int posLen = 8900;
    /// var grid = ToGrid(adList, posList, posOfGrid = 10, gridLen = posLen/10);
    /// 
    /// </summary>
    public class TimeGridAdvHelper
    {
        NLog.Logger logger = NLog.LogManager.GetCurrentClassLogger();
        
        /// <summary>
        /// ad数据时间间隔
        /// </summary>
        double ad_ts_ms = 1.28;

        #region 一分钟数据缓存池, 目标每个AD数据（1.28ms） 都有对应的其它数据
        /// <summary>
        /// AD数据池时间间隔为1.28ms, 肯定最少有一个数据 
        /// </summary>
        public List<DateTimeUnit3> ADPool = new List<DateTimeUnit3>();

        /// <summary>
        /// pos 数据池, 肯定最少有一个数据
        /// </summary>
        public List<DateTimeUnit> PosPool = new List<DateTimeUnit>();

        /// <summary>
        /// 输入端口 数据池, 肯定最少有一个数据
        /// </summary>
        public List<DateTimeUnit2> IStatusPool = new List<DateTimeUnit2>();

        /// <summary>
        /// pos2 数据池, 肯定最少有一个数据
        /// </summary>
        public List<DateTimeUnit> Pos2Pool = new List<DateTimeUnit>();

        /// <summary>
        /// DriveStatus 数据池, 肯定最少有一个数据
        /// </summary>
        public List<DateTimeUnit4> DriveStatusPool = new List<DateTimeUnit4>();
        #endregion

        DateTime NewestTime;
        /// <summary>
        /// 上一次查找istatus的序号
        /// </summary>
        int istatus_start_idx = -1;
        /// <summary>
        /// 上一次查找pos的序号
        /// </summary>
        int pos_start_idx = -1;
        int pos2_start_idx = -1;
        int driveStatus_start_idx = -1;

        void ResetSearch() 
        {
            istatus_start_idx = -1;
            pos_start_idx = -1;
            pos2_start_idx = -1;
            driveStatus_start_idx = -1;
        }
        /// <summary>
        /// 
        /// </summary>
        public void Init()
        {
            Clear();
        }



        /// <summary>
        /// 添加ad数据，且判断数据量是否够触发事件, 时间间隔为1.28ms
        /// </summary>
        /// <param name="end_dt"></param>
        /// <param name="adArray"></param>
        public void AddAD(DateTime end_dt, int[] adArray)
        {
            if (adArray.Count() == 0)
                return;
            //------------------------------------------------------------------------------------------
            //AD数据添加
            DateTime dt = DateTime.MinValue;
            ADPool.Add(new DateTimeUnit3() { dt = end_dt, datas = adArray});

            //long ticks = (long)((ts_ms * adArray.Count()) * TimeSpan.TicksPerMillisecond);
            NewestTime = dt;
            LimitPool();
        }
        void LimitPool() 
        {
            LimitSimplePool(ADPool);
            LimitSimplePool(PosPool);
            LimitSimplePool(Pos2Pool);
            LimitSimplePool(DriveStatusPool);
            LimitSimplePool(IStatusPool);
        }
        void LimitSimplePool<T>(List<T> pool) where T : IDataTimeUnit {
            var limitTime = NewestTime - TimeSpan.FromMinutes(1);
            //删除Pos 数据池
            int removeCnt = 0;
            for (int i = 0; i < pool.Count(); i++)
            {
                var cell = pool[i];
                if (cell.dt < limitTime)
                {
                    //需要删除
                    removeCnt = i + 1;
                }
                else
                {
                    break;
                }
            }

            if (removeCnt > 0)
            {
                //不能全部数据删除完,必须留一个

                if (removeCnt == pool.Count())
                {
                    removeCnt--;
                    if (removeCnt <= 0)
                        return;//不用删除了
                }

                pool.RemoveRange(0, removeCnt);
            }
        }

        /// <summary>
        /// 添加脉冲
        /// </summary>
        /// <param name="pos"></param>
        /// <param name="time"></param>
        public void AddPos(DateTime time, int pos)
        {
            AddSimpleData(PosPool, time, pos);
        }

        /// <summary>
        /// 添加驱动状态
        /// </summary>
        /// <param name="driveStatus"></param>
        /// <param name="time"></param>
        public void AddDriveStatus(DateTime time,DRIVE_MAN_ORDER order,  DRIVE_MAN_STATUS driveStatus, int marker)
        {
            var pool = DriveStatusPool;
            if (pool.Count > 0)
            {
                if (pool.Last().dt > time)
                {
                    //异常，把之前的数据删除
                    pool.Clear();
                }
                else if (pool.Last().dt == time)
                {
                    //删除之前那个
                    pool.RemoveAt(pool.Count() - 1);
                }
            }
            pool.Add(new DateTimeUnit4() { dt = time, order = order, driveStatus = driveStatus, marker = marker });
        }
        void AddSimpleData(List<DateTimeUnit> pool, DateTime time, int data) {

            if (pool.Count > 0)
            {
                if (pool.Last().dt > time)
                {
                    //异常，把之前的数据删除
                    pool.Clear();
                }
                else if (pool.Last().dt == time) 
                {
                    //删除之前那个
                    pool.RemoveAt(pool.Count() - 1);
                }
            }
            pool.Add(new DateTimeUnit() { dt = time, data = data });
        }
        
        /// <summary>
        /// 添加脉冲
        /// </summary>
        /// <param name="pos2"></param>
        /// <param name="time"></param>
        public void AddPos2(DateTime time, int pos2)
        {
            AddSimpleData(Pos2Pool, time, pos2);
        }

        /// <summary>
        /// 添加输入口状态
        /// </summary>
        /// <param name="time"></param>
        /// <param name="istatus"></param>
        /// <param name="pos"></param>
        /// <param name="pos2"></param>
        public void AddIStatus(DateTime time, UInt16 istatus, int pos, int pos2)
        {
            var pool = IStatusPool;
            if (pool.Count > 0)
            {
                if (pool.Last().dt >= time)
                {
                    //异常，把之前的数据删除
                    pool.Clear();
                }
            }

            IStatusPool.Add(new DateTimeUnit2() { dt = time, istatus = istatus, isPosValid = true,  pos = pos, pos2 = pos2 });
            AddPos(time, pos);
            AddPos2(time, pos2);
        }

        /// <summary>
        /// 添加开始位置
        /// </summary>
        /// <param name="pos"></param>
        /// <param name="time"></param>
        public void AddPos_Default(DateTime time, int pos)
        {
            if (PosPool.Count() == 0)
            {
                AddPos(time,pos);
            }
        }
        
        /// <summary>
        /// 添加开始位置
        /// </summary>
        /// <param name="pos"></param>
        /// <param name="time"></param>
        public void AddPos2_Default(DateTime time, int pos)
        {
            if (Pos2Pool.Count() == 0)
            {
                AddPos2(time,pos);
            }
        }


        /// <summary>
        /// 添加开始驱动状态
        /// </summary>
        /// <param name="time"></param>
        /// <param name="order"></param>
        /// <param name="dirveStatus"></param>
        /// <param name="marker"></param>
        public void AddDriveStatus_Default(DateTime time, DRIVE_MAN_ORDER order, DRIVE_MAN_STATUS dirveStatus, int marker)
        {
            if (IStatusPool.Count() == 0)
            {
                AddDriveStatus(time, order, dirveStatus, marker);
            }
        }

        /// <summary>
        /// 添加开始状态
        /// </summary>
        /// <param name="istatus"></param>
        /// <param name="time"></param>
        public void AddIStatus_Default(DateTime time, UInt16 istatus)
        {
            if (IStatusPool.Count() == 0)
            {
                IStatusPool.Add(new DateTimeUnit2() { dt = time, istatus = istatus});
            }
        }
        /// <summary>
        /// 
        /// </summary>
        public void Clear()
        {
            PosPool.Clear();
            ADPool.Clear();
            IStatusPool.Clear();
            Pos2Pool.Clear();
            DriveStatusPool.Clear();
            ResetSearch();
        }


        /// <summary>
        /// 获取最后一次运行的时间段
        /// </summary>
        /// <param name="begin"></param>
        /// <param name="end"></param>
        /// <param name="marker"></param>
        /// <param name="order"></param>
        /// <returns></returns>
        public bool GetLastRunningTime(out DateTime begin, out DateTime end, out int marker, out DRIVE_MAN_ORDER order) {
            begin = DateTime.MinValue;
            end = DateTime.MinValue;
            marker = 0;
            order = DRIVE_MAN_ORDER.IDLE;
            //1. 找 最后一个 driveStatus!=running的
            int idxOfStop = -1;
            for (int i = DriveStatusPool.Count() - 1; i >= 0; i--)
            {

                var cell = DriveStatusPool[i];
                if (cell.driveStatus != DRIVE_MAN_STATUS.RUNNING)
                {
                    //对！！它是。但继续向前找找看
                    idxOfStop = i;
                }
                else
                {
                    if (idxOfStop != -1)
                    {
                        //之前已经找到了，
                        break;
                    }
                }
            }
            if (idxOfStop == -1)
            {
                //一直在running，还没停下来
                return false;
            }

            //2. 最向前找running的
            int idxOfRunning = -1;
            for (int i = idxOfStop - 1; i >= 0; i--)
            {
                var cell = DriveStatusPool[i];
                if (cell.driveStatus == DRIVE_MAN_STATUS.RUNNING)
                {
                    //对！！它是。但继续向前找找看
                    idxOfRunning = i;
                }
                else
                {
                    //没了
                    break;
                }
            }

            if (idxOfRunning == -1)
            {
                //一直停着，没有运行过
                return false;
            }

            begin = DriveStatusPool[idxOfRunning].dt;
            end = DriveStatusPool[idxOfStop].dt;
            order = DriveStatusPool[idxOfStop].order;
            marker = DriveStatusPool[idxOfStop].marker;
            return true;
        }
        
        /// <summary>
        /// 获取最后一次运行中的timegrid;
        /// 找到最后一个 driveStatus!=running的, 最向前找running的
        /// </summary>
        /// <returns></returns>
        List<DateTimeUnit3> GetTimeGrid2(DateTime begin, DateTime end) 
        {
            //获取 从 idxOfRunning 到 idxOfStop 时间段的 timegrid
            List<DateTimeUnit3> reponse = new List<DateTimeUnit3>();
            for (int i = 0; i < ADPool.Count(); i++) {
                var cell = ADPool[i];
                if (cell.dt < begin)
                    continue;
                else if (cell.dt <= end)
                    reponse.Add(cell);
                else
                    break;
            }
            return reponse;
        }


        /// <summary>
        /// 获取最后一次运行中的timegrid;
        /// 找到最后一个 driveStatus!=running的, 最向前找running的
        /// </summary>
        /// <param name="begin">开始时间</param>
        /// <param name="end">结束时间</param>
        /// <param name="reponse_endTime">返回数据的最后一个数据的时间，数据间隔为1.28ms</param>
        /// <returns></returns>
        public List<int> GetTimeGrid(DateTime begin, DateTime end, out DateTime reponse_endTime)
        {
            reponse_endTime = DateTime.MinValue;
            List<int> reponse = new List<int>();

            long ticks = (long)(ad_ts_ms * TimeSpan.TicksPerMillisecond);
            var ts = TimeSpan.FromTicks(ticks);
            var reponse2 = GetTimeGrid2(begin, end);
            if (reponse2.Count() == 0)
                return reponse;


            DateTime lastTime = reponse2.Last().dt;
            reponse_endTime = lastTime;

            for (int i = 0; i < reponse2.Count(); i++) {
                int idx = reponse2.Count() - 1 - i;
                for (int j = 0; j < reponse2[idx].datas.Count(); j++) {
                    int idx2 = reponse2[idx].datas.Count() - 1 - j;
                    int ad = reponse2[idx].datas[idx2];
                    reponse.Add(ad);
                    lastTime -= ts;
                    if (lastTime < begin)
                        goto _end;//完成
                }
            }
            _end:
            //反转
            reponse.Reverse();
            return reponse;
        }
        
        public List<int> GetPos(DateTime end_dt, int cnt) {

            long ticks = (long)(ad_ts_ms * TimeSpan.TicksPerMillisecond);
            var ts = TimeSpan.FromTicks(ticks);

            List<int> reponse = new List<int>();

            DateTime lastTime = end_dt;
            for(int i=0;i<cnt;i++)
            {
                reponse.Add(GetPos(lastTime));
                lastTime -= ts;
            }
            //反转
            reponse.Reverse();
            return reponse;
        }


        public List<int> GetPos2(DateTime end_dt, int cnt)
        {
            long ticks = (long)(ad_ts_ms * TimeSpan.TicksPerMillisecond);
            var ts = TimeSpan.FromTicks(ticks);

            List<int> reponse = new List<int>();
            DateTime lastTime = end_dt;
            for (int i = 0; i < cnt; i++)
            {
                reponse.Add(GetPos2(lastTime));
                lastTime -= ts;
            }
            //反转
            reponse.Reverse();
            return reponse;
        }

        public List<UInt16> GetIStatus(DateTime end_dt, int cnt)
        {
            long ticks = (long)(ad_ts_ms * TimeSpan.TicksPerMillisecond);
            var ts = TimeSpan.FromTicks(ticks);

            List<UInt16> reponse = new List<UInt16>();
            DateTime lastTime = end_dt;
            for (int i = 0; i < cnt; i++)
            {
                reponse.Add(GetIStatus(lastTime));
                lastTime -= ts;
            }
            //反转
            reponse.Reverse();
            return reponse;
        }

        /// <summary>
        /// 给定结束时间点,开始grid, 找到pos数据池 对应的开始序号 与 结束序号
        /// </summary>
        /// <param name="start_pos">开始脉冲</param>
        /// <param name="endtime">结束时间点</param>
        /// <param name="start_idx">输出 pos数据池开始序号</param>
        /// <param name="end_idx">输出 pos数据池结束序号</param>
        /// <returns>false 发生在数据池前面， true成功获取</returns>
        bool GetIdxFromPosPool(int start_pos, DateTime endtime, out int start_idx, out int end_idx)
        {
            start_idx = -1;
            end_idx = searchIndex(PosPool, endtime, -1);
            if (end_idx == -1)
            {
                //endtime 发生在过去很久以前
                return false;
            }

            int pos_last = int.MinValue;
            int i_last = -1;
            for (int i = end_idx; i >= 0; i--)
            {

                int pos = PosPool[i].data;

                //if (pos == start_pos)//找到了！！！！
                //{
                //    pos_last = pos;
                //    continue;
                //}

                if (pos_last == int.MinValue)
                {
                    pos_last = pos;
                    i_last = i;
                    continue;
                }

                if (pos_last == pos)
                {
                    continue;
                }

                if (Math.Abs(pos - start_pos) > Math.Abs(pos_last - start_pos))
                {
                    //越找越远，
                    //上一个就是了
                    start_idx = i_last;
                    return true;
                }
                else if (Math.Abs(pos - start_pos) == Math.Abs(pos_last - start_pos))
                {
                    //跨过了
                    start_idx = i_last;
                    return true;
                }

                pos_last = pos;
                i_last = i;
            }
            //找完
            start_idx = 0;
            return true;
        }



        /// <summary>
        /// 根据时间点，获取输入口状态
        /// </summary>
        /// <param name="dt"></param>
        /// <returns></returns>
        UInt16 GetIStatus(DateTime dt)
        {
            int index = searchIndex(IStatusPool, dt, istatus_start_idx);
            istatus_start_idx = index;

            if (index < 0)
                index = 0;
            
            return (UInt16)IStatusPool[index].istatus;
        }

        /// <summary>
        /// 根据时间点，获取输入口状态
        /// </summary>
        /// <param name="dt"></param>
        /// <param name="index"></param>
        /// <returns></returns>
        UInt16 GetIStatus(DateTime dt, out int index)
        {
            index = searchIndex(IStatusPool, dt, istatus_start_idx);
            istatus_start_idx = index;

            if (index < 0)
                return 0xffff;
            else
                return (UInt16)IStatusPool[index].istatus;
        }



        /// <summary>
        /// 以 [idx，idx+1] 的比例， 给出 time 计算 pos 
        /// </summary>
        /// <param name="posPool"></param>
        /// <param name="time"></param>
        /// <param name="idx"></param>
        /// <returns></returns>
        int GetPos(List<DateTimeUnit> posPool, DateTime time, int idx)
        {
            if (posPool.Count() <= 0)//不应该存在
            {
                throw new Exception("没数据，还没准备好，不能获取");
            }

            if (posPool.Count() == 1)
                return posPool.First().data;

            if (idx < 0)
                idx = 0;

            if (idx >= posPool.Count()-1)
                idx = posPool.Count() - 2;
            var cell0 = posPool[idx];
            var cell1 = posPool[idx + 1];
            double d_pos = cell0.data - cell1.data;
            if (d_pos == 0)
                return cell0.data;
            
            TimeSpan ts = cell0.dt - cell1.dt;
            double p = d_pos * (time - cell1.dt).Ticks / ts.Ticks;

            return (int)(p + cell1.data);
        }

        enum SearchState 
        {
            ContinueLastTime,
            Back,
            ForwToCheck
        }
        /// <summary>
        /// 找到 时间 dt 落在 [idx，idx+1] 之间的 idx;
        /// 返回 -1，证明dt 在列表在前面， 或列表数据量为=0,
        /// 当dt 大于 列表最后一个数， 返回最后一个数据的id
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="list">列表</param>
        /// <param name="dt"></param>
        /// <param name="searchStartIdx">从列表的某个位置开始向前查找</param>
        /// <returns></returns>
        int searchIndex<T>(List<T> list, DateTime dt, int searchStartIdx ) where T: IDataTimeUnit
        {
            if (list.Count() == 0)
                return -1;

            int max_index = list.Count() - 1;

            if (dt >= list[max_index].dt)
                return max_index;

            if (dt < list[0].dt)
                return -1;


            int index = searchStartIdx;
            SearchState state = SearchState.ContinueLastTime;

            if ((index < 0) || (index >= max_index))//重新开始
            {
                index = max_index;
                state = SearchState.Back;
            }
            else if (state == SearchState.ContinueLastTime) 
            {
                if (list[index].dt <= dt)
                {
                    //可能找到了
                    //再向正方向看看
                    state = SearchState.ForwToCheck;
                }
                else
                {
                    state = SearchState.Back;
                }
            }

            if (state == SearchState.ForwToCheck)
            {
                while (true) 
                {
                    DateTime time = list[index].dt;
                    if (time > dt)
                    {
                        //就是之前那个
                        index--;
                        return index;
                    }
                    else
                    {
                        index++;
                    }
                    if (index > max_index)
                    {
                        //最后一个就是了
                        throw new Exception("之前没有判断出 时间点在列表后，异常！！");
                        //return max_index;
                    }
                }
            }
            else//SearchState.Back
            {
                while (true)
                {
                    DateTime time = list[index].dt;
                    if (time <= dt)
                    {
                        //就是它
                        return index;
                    }
                    index--;
                    if(index<0)
                    {
                        //不可能发生
                        throw new Exception("之前没有判断出 时间点在列表前，异常！！");
                    }
                }
            }
        }


        /// <summary>
        /// 获取时间对应的 位置点
        /// </summary>
        /// <param name="dt"></param>
        /// <returns></returns>
        int GetPos(DateTime dt)
        {
            //找到 时间落在 [idx，idx+1] 之间的 序号
            int index = searchIndex(PosPool, dt, pos_start_idx);
            pos_start_idx = index;

            if (index < 0) 
                index = 0;

            return GetPos(PosPool, dt, index);
        }
        int GetPos2(DateTime dt) 
        {
            //找到 时间落在 [idx，idx+1] 之间的 序号
            int index = searchIndex(Pos2Pool, dt, pos2_start_idx);
            pos2_start_idx = index;

            if (index < 0)
                index = 0;

            return GetPos(Pos2Pool, dt, index);
        }
       
        

        void toGrid_step1(int[] grids_sum, int[] grids_cnt, int gridLen, int grid_idx, int ad) 
        {
            if ((grid_idx < 0) || (grid_idx >= gridLen))
            {
                return;
            }

            if (!Misc.MyBase.ISVALIDATA(ad))
                return;

            grids_cnt[grid_idx]++;
            grids_sum[grid_idx] += ad;
        }

        public int[] ToGrid(List<int> adList, List<int> posList, int posOfGrid, int gridLen) 
        {
            int[] grids_sum = new int[gridLen];
            int[] grids_cnt = new int[gridLen];

            int grid_idx_last = -1;
            //grid_idx 应该要连续变化，不是，那就插值
            for (int j = 0; j < adList.Count() && j<posList.Count(); j++)
            {
                int ad = adList[j];
                int pos = posList[j];
                int grid_idx = pos / posOfGrid;
                if (grid_idx_last != -1 && (Math.Abs(grid_idx_last - grid_idx) > 1))
                {
                    //步子太大，需要插值
                    if (grid_idx_last < grid_idx)
                    {
                        for (int i = grid_idx_last; i < grid_idx; i++)
                        {
                            toGrid_step1(grids_sum, grids_cnt, gridLen, i, ad);
                        }
                    }
                    else
                    {
                        for (int i = grid_idx_last; i > grid_idx; i--)
                        {
                            toGrid_step1(grids_sum, grids_cnt, gridLen, i, ad);
                        }
                    }
                }
                toGrid_step1(grids_sum, grids_cnt, gridLen, grid_idx, ad);
                grid_idx_last = grid_idx;
            }

            for (int i = 0; i < gridLen; i++)
            {
                if (grids_cnt[i] == 0)
                    grids_sum[i] = Misc.MyBase.NULL_VALUE;
                else
                    grids_sum[i] = (int)(grids_sum[i] / grids_cnt[i]);
            }
            return grids_sum;
        }

        public void GetIStatusRange(
            int istatusIndex,
            DateTime beginTime, DateTime endTime,
            out List<Range> data0, out List<Range> data1)
        {

            data0 = new List<Range>();
            data1 = new List<Range>();
            UInt16 istatus = GetIStatus(beginTime, out int istatus_idx);
            int beginPos = GetPos(beginTime);
            int endPos = GetPos(endTime);
            DIRECTION direction = endPos > beginPos ? DIRECTION.FORWARD : DIRECTION.BACKWARD;
            if (istatus_idx == -1)//异常，列表没有数据
            {
                //全部都是
                if (direction == DIRECTION.FORWARD)
                    data1.Add(new Range() { Begin = beginPos, End = endPos });
                else
                    data1.Add(new Range() { Begin = endPos, End = beginPos });

                return;
            }
            bool istatus_no_b = Misc.MyBase.CHECKBIT(istatus, istatusIndex);

            Range r = new Range();

            r.Begin = beginPos;

            for (int i = istatus_idx + 1; i < IStatusPool.Count(); i++)
            {
                istatus = IStatusPool[i].istatus;
                int pos = IStatusPool[i].pos;
                bool b = Misc.MyBase.CHECKBIT(istatus, istatusIndex);
                if (IStatusPool[i].dt >= endTime)
                {
                    //结束，这是最后！！！
                    break;
                }
                else
                {
                    if (b != istatus_no_b)
                    {
                        r.End = pos;
                        if (istatus_no_b)
                            data1.Add(r);
                        else
                            data0.Add(r);
                        istatus_no_b = b;
                        r = new Range();
                        r.Begin = pos;
                    }
                }
            }
            r.End = beginPos;
            if (istatus_no_b)
                data1.Add(r);
            else
                data0.Add(r);

            if (direction == DIRECTION.FORWARD)
            {
                for (int i = 0; i < data0.Count(); i++)
                {
                    data0[i].End -= 1;
                }
                for (int i = 0; i < data1.Count(); i++)
                {
                    data1[i].End -= 1;
                }
            }
            else
            {
                for (int i = 0; i < data0.Count(); i++)
                {

                    data0[i].End += 1;
                    int t = data0[i].End;
                    data0[i].End = data0[i].Begin;
                    data0[i].Begin = t;
                }
                data0.Reverse();


                for (int i = 0; i < data1.Count(); i++)
                {
                    data1[i].End += 1;
                    int t = data1[i].End;
                    data1[i].End = data1[i].Begin;
                    data1[i].Begin = t;
                }
                data1.Reverse();
            }
        }

    }
    public interface IDataTimeUnit 
    {
        DateTime dt { get; set; }
    }
    public struct DateTimeUnit: IDataTimeUnit
    {
        public int data;
        public DateTime dt { get; set; }
        public override string ToString()
        {
            return $"{dt.Ticks} | {data}";
        }
    }

    public struct DateTimeUnit4 : IDataTimeUnit
    {
        public DateTime dt { get; set; }
        public DRIVE_MAN_ORDER order;
        public DRIVE_MAN_STATUS driveStatus;
        public int marker;

        public override string ToString()
        {
            return $"{dt.Ticks} | {order} | {driveStatus} | {marker}";
        }
    }
    public struct DateTimeUnit3 : IDataTimeUnit
    {
        /// <summary>
        /// 数据结束点
        /// </summary>
        public DateTime dt { get; set; }

        /// <summary>
        /// 数据列
        /// </summary>
        public int[] datas;
        
        public override string ToString()
        {
            return $"{dt.Ticks} | cnt={datas.Count()}";
        }
    }

    public struct DateTimeUnit2: IDataTimeUnit
    {
        public DateTime dt { get; set; }
        public UInt16 istatus;
        public bool isPosValid;
        public int pos;
        public int pos2;
        public override string ToString()
        {
            if(isPosValid)
                return $"{dt.Ticks} | ({istatus:X2}) | 1={pos} | 2={pos2}";
            else
                return $"{dt.Ticks} | ({istatus:X2})";

        }
    }
}
